from __future__ import annotations

import shlex
import shutil
import sys
import asyncio
# from codecs import ignore_errors
# from itertools import zip_longest
# from operator import itemgetter
# from re import search
from subprocess import PIPE
from dataclasses import dataclass

import click
# from rich import inspect
from rich.text import Text, TextType
# from textual._two_way_dict import TwoWayDict
from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, VerticalScroll, Vertical
from textual.reactive import reactive
from textual.screen import ModalScreen, Screen
from textual.widget import Widget
from textual.widgets import Header, Static, RichLog, DataTable, Footer, HelpPanel, Markdown, Sparkline, Input
from textual import work, events, on
from textual.message import Message
from textual.binding import Binding
from textual.actions import SkipAction
# from textual.widgets._data_table import ColumnKey, CellType, RowKey, CellDoesNotExist, Row

from mmng_ui.reader import ParseLine, PocsagMessage
from mmng_ui._version import __version__



@dataclass
class OutputMessage(Message, bubble=False):
    """Custom message class to handle subprocess output."""

    output: str


@dataclass
class Status:
    """The status pane"""

    receiver: str
    ip_address: str
    json_mode: bool
    charset: str

    def __repr__(self):
        return f'Receiver: {self.receiver}\nIP address: {self.ip_address}'


class UDPHandler(asyncio.DatagramProtocol):
    """Handle UDP traffic"""

    def __init__(self, app, loop):
        self.app = app
        self.loop = loop
        self.status = self.app.query_one('#status')
        self.last_activity_time = 0

    def connection_made(self, transport):
        self.transport = transport
        self.status.receiver = 'ready'

    def connection_lost(self, exc):
        self.status.receiver = 'Closed'

    def datagram_received(self, data, addr):
        self.last_activity_time = self.loop.time()
        self.status.ip_address = addr[0]
        self.app.process.stdin.write(data)
        self.app.process.stdin.drain()

    async def idle_task(self):
        """This updates the things in the status pane."""
        while True:
            if self.loop.time() - self.last_activity_time > 5:
                self.status.receiver = '[wheat4]idle[/]'
            await asyncio.sleep(1)


class StatusWidget(Widget):
    """The status pane."""

    receiver = reactive('[dark_red]Not connected[/]')
    ip_address = reactive('[wheat4]None[/]')
    json_mode = reactive('[wheat4]Unknown[/]')
    charset = reactive('[wheat4]Unknown[/]')

    def render(self) -> str:
        return f'Receiver: {self.receiver}\nIP address: {self.ip_address}\nJSON mode: {self.json_mode}\nCharset: {self.charset}'


class HelpScreen(ModalScreen):
    """Help screen modal."""

    BINDINGS = [("escape,space,q,question_mark", "app.pop_screen", "Close")]

    def compose(self) -> ComposeResult:
        text = """
# mmng-ui

## Introduction

This is a TUI utility to decode and see POCSAG messages.

mmng-ui will listen on a chosen UDP port for raw streams from software like SDR++, use multimon-ng to decode it, and
show you POCSAG messages in a wonderful text interface.

## Usage

Alpha POCSAG messages will display in the top pane.  The bottom pane will show the raw output from `multimon-ng`,
as well as any errors or issues with decoding.

The status panel shows any incoming connections.  Receiver will transition between the following states:

| Receiver state | Description                                     |
|----------------|-------------------------------------------------|
| idle           | No UDP traffic yet seen, or seen in 5 seconds   |
| receiving      | Actively receiving a decode from `multimon-ng`  |
| waiting        | Traffic is coming in, but nothing to be decoded |

Just below the status panel is a sparkline -- this updates on each decode, and reflects character length of said
decode.

Underneath the log window in another sparkline, and this shows messages per second, for the last minute.

The footer shows available keyboard choices to quit the app, show a help screen, and clear all logging panes.

The mouse will also work!

## JSON mode

`mmng-ui` will attempt to auto-detect the output format from `multimon-ng`, and if it looks like JSON, it'll use it.

JSON output isn't yet in `multimon-ng`, but I have a working
fork [here](https://github.com/lingfish/multimon-ng/tree/add-json).

[//]: # (README.md ends here)"""
        yield Markdown(text, id='help')


# class FilterScreen(ModalScreen[str]):
#     """Screen with a dialog to quit."""
#
#     BORDER_TITLE = 'Filter messages'
#
#     def compose(self) -> ComposeResult:
#         yield Input(placeholder='Enter a filter here', id='filter')
#
#     @on(Input.Submitted)
#     def handle_filter(self, event: Input.Submitted) -> None:
#         self.dismiss(event.value)


class MsgsPerSecond(Sparkline):
    """Calculate/update the messages per second sparkline."""

    def __init__(self, samples=[0] * 60, **kwargs):
        super().__init__(**kwargs)
        self.samples = samples

    def on_mount(self) -> None:
        self.update_timer = self.set_interval(1, self.update_graph)
        self.data = self.samples

    def update_graph(self) -> None:
        self.data = self.data[-59:] + [len(self.app.message_count)]
        self.app.message_count = []


# class DataTableFilter(DataTable):
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self._unfiltered_data = None
#         self._unfiltered_rows = None
#         self._search_term = None
#         self._search_column = None
#
#     async def filter(self, column: str, search: str) -> Self:
#         self._search_term = search
#         self._search_column = column
#         if search == '':
#             if self._unfiltered_data is not None:
#                 self._data = self._unfiltered_data
#                 self._unfiltered_data = None
#
#                 self.rows = self._unfiltered_rows
#                 self._unfiltered_rows = None
#                 self.border_title = 'POCSAG messages'
#         else:
#             if self._unfiltered_data is None:
#                 self._unfiltered_data = self._data
#                 self._unfiltered_rows = self.rows
#
#             self._data = dict(
#                 filter(
#                     lambda x: True if search.lower() in str(x[1][column]).lower() else False,
#                     self._unfiltered_data.items(),
#                 )
#             )
#             self.rows = {row_key: self._unfiltered_rows[row_key] for row_key in self._data.keys()}
#             self.border_title = 'POCSAG messages (filter applied)'
#
#         self._row_locations = TwoWayDict({key: new_index for new_index, (key, _) in enumerate(self._data.items())})
#         self._update_count += 1
#         self._require_update_dimensions = True
#         self.refresh()
#         return self
#
#     @property
#     def filter_active(self) -> bool:
#         return self._search_term != ''
#
#     async def filter_refresh(self) -> None:
#         if self.filter_active:
#             await self.filter(self._search_column, '')
#             await self.filter(self._search_column, self._search_term)
#
#     def update_cell(
#         self, row_key: RowKey | str, column_key: ColumnKey | str, value: CellType, *, update_width: bool = False
#     ) -> None:
#         if self._unfiltered_data is not None:
#             try:
#                 self._unfiltered_data[row_key][column_key] = value
#             except KeyError:
#                 raise CellDoesNotExist(f'No cell exists for row_key={row_key!r}, column_key={column_key!r}.') from None
#
#         return super().update_cell(row_key, column_key, value, update_width=update_width)
#
#     def clear(self, columns: bool = False) -> Self:
#         self._unfiltered_data = None
#         self._unfiltered_rows = None
#         return super().clear(columns)
#
#     def add_column(
#         self, label: TextType, *, width: int | None = None, key: str | None = None, default: CellType | None = None
#     ) -> ColumnKey:
#         column_key = super().add_column(label, width=width, key=key, default=default)
#
#         if self._unfiltered_data is not None:
#             for row_key in self._unfiltered_rows.keys():
#                 self._unfiltered_data[row_key][column_key] = default
#
#         return column_key
#
#     def add_row(
#         self, *cells: CellType, height: int = 1, key: str | None = None, label: TextType | None = None
#     ) -> RowKey:
#         row_key = super().add_row(*cells, height=height, key=key, label=label)
#
#         if self._unfiltered_data is not None:
#             self._unfiltered_data[row_key] = {
#                 column.key: cell for column, cell in zip_longest(self.ordered_columns, cells)
#             }
#             label = Text.from_markup(label) if isinstance(label, str) else label
#             self._unfiltered_rows[row_key] = Row(row_key, height, label)
#
#         return row_key
#
#     def remove_row(self, row_key: RowKey | str) -> None:
#         super().remove_row(row_key)
#         if self._unfiltered_data is not None:
#             del self._unfiltered_rows[row_key]
#             del self._unfiltered_data[row_key]
#
#     def remove_column(self, column_key: ColumnKey | str) -> None:
#         super().remove_column(column_key)
#
#         if self._unfiltered_data is not None:
#             for row in self._unfiltered_data:
#                 del self._unfiltered_data[row][column_key]


class MainScreen(Screen):
    def compose(self):
        yield Header()
        with Container(id='app-grid'):
            yield DataTable(id='messages')
            yield RichLog(id='log', highlight=True, markup=True)
            with Container(id='status-container'):
                yield StatusWidget(id='status')
                yield Sparkline([], id='spark')
        yield MsgsPerSecond(id='msgs-per-second')
        yield Footer()

    async def on_mount(self) -> None:
        """Setup the initial components."""
        self.current_width = '0'
        self.title = 'multimon-ng decoder'
        table = self.screen.query_one('#messages')
        log = self.screen.query_one('#log')
        status = self.screen.query_one('#status')

        table.add_column('Time', key='time')
        table.add_column('Address', key='address')
        table.add_column('Message', key='message')
        table.cursor_type = 'none'
        table.border_title = 'POCSAG messages'
        log.border_title = 'Log window'
        status.border_title = 'Status'

        self.parse_line = ParseLine()

        # Run multimon-ng, grab version and JSON support
        mmng_help_process = await asyncio.create_subprocess_exec(self.app.mmng_binary, '-h', stderr=PIPE)
        mmng_help = await mmng_help_process.stderr.read()
        await mmng_help_process.wait()
        mmng_text = mmng_help.decode()
        json_capable = '--json' in mmng_text
        log.write(f'multimon-ng version: {mmng_text.splitlines()[0]}')
        log.write(f'JSON capable: {json_capable}')

        status.charset = self.app.charset
        mmng_args = f'-a POCSAG512 -a POCSAG1200 -a POCSAG2400 -a FLEX -a FLEX_NEXT -f alpha -t raw -u -q --timestamp -p {"--json" if json_capable else ""} -C {self.app.charset} -'
        self.log('About to start multimon')
        self.stream_subprocess(self.app.mmng_binary, mmng_args)
        self.log('AFTER: About to start multimon')

    @work(exclusive=True)
    async def stream_subprocess(self, command, args):
        """Stream output from a subprocess and post it using post_message."""
        self.log('   in stream_subprocess')
        self.process = await asyncio.create_subprocess_exec(
            command, *shlex.split(args), stdin=PIPE, stdout=PIPE, stderr=PIPE
        )
        self.log('*** process is assigned')

        network_loop = asyncio.get_running_loop()
        transport, protocol = await network_loop.create_datagram_endpoint(
            lambda: UDPHandler(self, network_loop), local_addr=('::', self.app.port)
        )
        network_loop.create_task(protocol.idle_task())

        # Stream stdout asynchronously
        async for line in self.read_process_output(self.process.stdout):
            self.log(f'Raw output from multimon: {line}')
            self.post_message(OutputMessage(line))
            self.set_timer(1, lambda: setattr(self.query_one('#status'), 'receiver', '[dark_green]waiting[/]'))
            self.query_one('#spark').data = self.query_one('#spark').data[-9:] + [len(line)]
            self.app.message_count.append(1)

        # Handle any stderr errors
        async for error in self.read_process_output(self.process.stderr):
            self.post_message(OutputMessage(f'[red]Error: {error}'))

    async def read_process_output(self, output):
        """Read the output of a subprocess line by line."""
        self.log('   in read_process_output')
        status = self.query_one('#status')
        while True:
            status.receiver = '[blink bold bright_green]receiving[/]'
            line = await output.readline()
            self.log('   read a line')
            if not line:
                break
            yield line.decode().strip()

    async def on_resize(self, event: events.Resize) -> None:
        self.current_width = event.size.width
        table = self.screen.query_one('#messages')
        self.recalc_width(table)

    async def watch_show_vertical_scrollbar(self) -> None:
        table = self.screen.query_one('#messages')
        self.recalc_width(table)

    async def on_output_message(self, message: OutputMessage):
        """Handle OutputMessage to update UI components."""
        log = self.screen.query_one('#log')
        table = self.screen.query_one('#messages')
        status = self.screen.query_one('#status')

        self.log(f'RECEIVED EVENT: {message}')
        # Process the output as it becomes available
        log.write(f'[bold magenta]multimon-ng: {message.output}')

        result, json_detected = self.parse_line.parse(message.output)
        self.log(f'result: {result}')

        status.json_mode = json_detected

        if message and result.trim_message:
            self.log('Adding a row')
            table.add_row(
                str(result.current_time.strftime('%H:%M:%S')),
                Text(str(result.address), justify='right'),
                result.trim_message,
                height=None,
            )
            # await table.filter_refresh()
        else:
            log.write('WARNING: No valid message decoded from multimon-ng')

        self.recalc_width(table)

    def recalc_width(self, table) -> None:
        message_col_width = table.columns['time'].get_render_width(table) + table.columns['address'].get_render_width(
            table
        )
        if table.show_vertical_scrollbar:
            scroll_padding = table.styles.scrollbar_size_vertical
        else:
            scroll_padding = 0
        table.columns['message'].width = (
            (table.size.width - message_col_width) - (2 * table.cell_padding) - scroll_padding
        )
        table.columns['message'].auto_width = False
        try:
            table.action_scroll_bottom()
        except SkipAction:
            pass


class Pocsag(App):
    def __init__(self, mmng_binary: str, port: int, charset) -> None:
        self.mmng_binary = mmng_binary
        self.port = port
        self.charset = charset
        # self.filter: str = None
        super().__init__()

    CSS_PATH = 'pocsag.tcss'

    # SCREENS = {'help': HelpScreen, 'filter': FilterScreen}
    SCREENS = {'help': HelpScreen}

    BINDINGS = [
        Binding(key='q', action='quit', description='Quit the app'),
        Binding(
            key='question_mark',
            action="app.push_screen('help')",
            description='Show help screen',
            key_display='?',
        ),
        Binding(key='c', action='clear_screen', description='Clear all panes'),
        # Binding(key='/', action='filter', description='Filter the messages'),
    ]

    message_count = []

    def on_mount(self):
        self.push_screen(MainScreen())

    def action_clear_screen(self) -> None:
        self.screen.query_one('#messages').clear()
        self.screen.query_one('#log').clear()
    #
    # async def action_filter(self) -> None:
    #     async def check_filter(filter: str | None) -> None:
    #         """Called when FilterScreen is dismissed."""
    #         table = self.screen.query_one('#messages')
    #         self.log(filter)
    #         self.filter = filter
    #         await table.filter('message', search=self.filter)
    #
    #     await self.push_screen(FilterScreen(), check_filter)


@click.command(context_settings={'show_default': True})
@click.option('--mmng-binary', '-m', required=False, default='multimon-ng', help='Path to multimon-ng binary')
@click.option('--port', '-p', required=False, type=int, default=8888, help='Port to listen on')
@click.option(
    '--charset',
    '-c',
    type=click.Choice(['US', 'FR', 'DE', 'SE', 'SI'], case_sensitive=False),
    required=False,
    default='US',
    help='Charset encoding (case sensitive!)',
)
@click.version_option(version=__version__)
def main(mmng_binary, port, charset):
    if not shutil.which(mmng_binary):
        click.echo('multimon-ng binary not found!', err=True)
        sys.exit(1)

    Pocsag(mmng_binary, port, charset).run()


if __name__ == '__main__':
    main()



import warnings
from xml.sax import parse

import moment
import pytest

from mmng_ui.reader import ParseLine, PocsagMessage


FLEX_LINES = """FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [4294947544] UNK 00136080 001794b2 00193e1c 0015db79 0004e703 000136bd 00194da5 001b8f04 001846e3 001e8f83 001d7d3b 000ae060 00026bc2 001ffff3 001421c8
FLEX|2024-12-08 21:18:16|3200/4/F/D|03.096|002064197|ALN|
FLEX|2024-12-08 21:18:16|3200/4/C/D|03.096|4294956915|ALN|A - 16 NORTH 41  [34]
FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [002064209] NUM U75[
FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [4294964846] UNK 00116080 000f8b51 000b60aa 000f86aa 0018a645 001620c1 001c1132 000179e8 0002a489 001e10da 0007e16a 001bb879 0009db37 00000008 001421c3
FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [002064201] NUM 73798 94U
FLEX|2024-12-08 21:18:16|3200/4/C/D|03.096|4294947155|ALN|A - 16 NORTH 41  [39]
FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [002064225] NUM 838[
FLEX: 2024-12-08 21:18:16 3200/4/D 03.096 [4294950812] UNK 000a2080 00182222 00178935 0015f288 0010e2e4 000c4579 00111cb0 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
FLEX|2024-12-08 21:18:35|3200/4/C/A|04.062|4294941051|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [73]C9
FLEX: 2024-12-08 21:18:35 3200/4/A 04.062 [002064200] NUM 0-82
FLEX|2024-12-08 21:18:46|3200/4/C/A|04.063|4294947329|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [18]
FLEX: 2024-12-08 21:18:46 3200/4/A 04.063 [002064196] NUM U159
FLEX: 2024-12-08 21:18:46 3200/4/A 04.063 [4294962628] UNK 00000080 0007b3af 0001f034 00183d0a 001a02fd 001dfcf6 001b64de 00000002 001ca342
FLEX: 2024-12-08 21:18:46 3200/4/A 04.063 [002064198] TON 
FLEX|2024-12-08 21:18:46|3200/4/C/A|04.063|4294944497|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [45]
FLEX: 2024-12-08 21:18:46 3200/4/A 04.063 [002064200] NUM 0-82
FLEX: 2024-12-08 21:18:46 3200/4/B 04.063 [4294964639] UNK 00090080 0005b7fc 001fb6d5 00089a00 00162f2d 00019ed2 0019675d 001fa814 001c614d 001e1603 0013751e 0010e361 00096f78 001d1514 001020f0 0014f05e 0010f246 001f50c9 00116430 0012c654 000811bc 00013496 00181839 0006f87c 000c6a85 00025af3 0007af6b 000f88a8 0000b0b4 00147e3e 00025293 0000024e 00000000
FLEX|2024-12-08 21:18:46|3200/4/C/B|04.063|002064201|ALN|
FLEX|2024-12-08 21:18:46|3200/4/C/C|04.063|4294942723|ALN|nsult LOPEZ PEREZ 31373343 h/o spinal compression fx's MRI 12/7 subacute compression deformities, no evidence of cord compression. Pt neurologically intact. any acute surgical intervention? - Donald Thommes 6314176868 [68]3fL
FLEX|2024-12-08 21:18:46|3200/4/F/C|04.063|002064198|ALN|
FLEX: 2024-12-08 21:19:07 3200/4/A 04.064 [4294951837] UNK 001e0080 001b8c62 0004af2b 0017e3b9 000d7361 000d0cbd 0018afef 000eeadc 0013422d 001aafdb 000154e5 00012a9c 0019da58 00168443 000c8ff0 000ad2b5 0015fbba 000dad9a 000e1a3b 000168ad 00175e11 001ce511 000b7428 0002f90b 0017584e 001ce36c 001be44f 00189429 0001c4d5 001347e9 00000364 000002fc 00074080
FLEX: 2024-12-08 21:19:07 3200/4/A 04.064 [4294947930] UNK 00074080 00010ecb 00017eb4 000b534f 000c83e1 00194902 00086a68 001ffffc 001ca342
FLEX|2024-12-08 21:19:07|3200/4/C/A|04.064|002064199|ALN|
FLEX|2024-12-08 21:19:07|3200/4/C/A|04.064|4294954932|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [81]3f
FLEX: 2024-12-08 21:19:07 3200/4/A 04.064 [002064197] NUM 0208
FLEX|2024-12-08 21:19:07|3200/4/C/C|04.064|4294959999|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [68]EFr
FLEX: 2024-12-08 21:19:07 3200/4/C 04.064 [002064200] NUM 494
FLEX|2024-12-08 21:19:07|3200/4/C/C|04.064|4294954950|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [42]
FLEX: 2024-12-08 21:19:07 3200/4/C 04.064 [002064197] NUM [82
FLEX|2024-12-08 21:19:24|3200/4/C/A|04.065|4294942993|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [63]f
FLEX: 2024-12-08 21:19:24 3200/4/A 04.065 [002064255] NUM 0-82
FLEX|2024-12-08 21:20:10|3200/4/C/A|04.067|4294934942|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [74]n9q
FLEX: 2024-12-08 21:20:10 3200/4/A 04.067 [002064200] NUM 615 
FLEX|2024-12-08 21:20:10|3200/4/F/A|04.067|4294944338|ALN|y880kbOfI5gfYoG5q6BAA<Fr
FLEX|2024-12-08 21:20:10|3200/4/K/A|04.067|002064205|ALN|
FLEX|2024-12-08 21:20:10|3200/4/C/A|04.067|4294954940|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [69]
FLEX: 2024-12-08 21:20:10 3200/4/A 04.067 [002064197] NUM 0-82
FLEX|2024-12-08 21:20:10|3200/4/C/C|04.067|4294960016|ALN|om: meditechalerts@ehr.org Subject: Housekeeping Request - Dirty: 435 02  [34]
FLEX: 2024-12-08 21:20:10 3200/4/C 04.067 [002064200] NUM 1-82
FLEX|2024-12-08 21:20:35|3200/4/F/A|04.068|4294957717|ALN|Io5VcvE4axx96Gao5S27oVGTMCtWrYop7acr6bHwpuSkBVYy6mH1/hoGv90tkB56M31MDsDRYpr3UrZHtIzAagGC7wRUpYesyLcB7J4cjG4oTSN+8XtVaRJCcC//W/VLI9Wyew0wrwNfwU9LoQvqY7WDuToJZ0h1CNtTSCfwMjFgnn3f
FLEX: 2024-12-08 21:20:35 3200/4/A 04.068 [002064208] UNK
FLEX|2024-12-08 21:20:35|3200/4/C/B|04.068|4294949507|ALN|gjHWy4P7KiHaoSjJNDh5FFBLcrZUE6aH1ob2BjUDr9Bwzv9Hj3EFGTCC4e6ReNxVpUqrGjSTIuc2lkPv6wdf0KYcpIgY4JKE2nONeRhokfPl1kWqgR7tlhY0o2uXqkfsve1jyLQuiYCc/HXDuCMka0wrqgUEzdY354tXBSEpfuA8ZI
FLEX|2024-12-08 21:20:35|3200/4/K/C|05.068|4294958624|ALN|O2lH3hZo1Msa3QGxn9A8mh4xIxrhv1CZB2dnl6bZR1PTZLnqn1jhJyFezB+1C9+uWyHk7J4b2PI3FiSFkY8L/fBFEes27TOymNS79T1uqMTFrvWeRp/QL1HqWhYc5WaMBvCSdmGrj7kaY7B4oEu7BvzAkTiFqeV8+fxqyIEiYduRt3fL
FLEX: 2024-12-08 21:20:35 3200/4/C 05.068 [002064209] UNK
FLEX|2024-12-08 21:20:35|3200/4/F/D|05.068|4294948699|ALN|5fIIzRaReOOgQdqS2sCeu/vzt0ip8gQo+gf3oMhdcDDF/Y5H5f0/IlFaQ5G11QUhMemAYPtyUxe/Lahtxbw/QS1DB9SKk5Q971ZQvOsadK3mSbSMmDKCoEVH1MzbXeMRbUx2hP9YIShPEf7TuqlLpvADxWvcwcOaN6Goly2sEO95kP
FLEX: 2024-12-08 21:20:35 3200/4/D 05.068 [002064214] NUM 78-6
FLEX|2024-12-08 21:20:46|3200/4/C/A|04.069|4294957717|ALN|JAL9h6uvKnu86BtKZ3XpCIfW2uncjuItbKwgHTOVtpz33LXWm+LjOHgApI4P7C/aEM1MykbsDFCPgN60nL0jBdakgBS/RSWiilKgRjoplyJUWF46JrCFTm55brpVQCvTf2CELdy71L6q1aw33qQjUSPJigwqJB1yXWDAAmg3
FLEX|2024-12-08 21:20:46|3200/4/F/A|04.069|002064208|ALN|
FLEX: 2024-12-08 21:20:46 3200/4/A 04.069 [002064209] NUM 358
"""


@pytest.fixture
def sample_data():
    return '2024-09-23 12:38:00: POCSAG512: Address:  162202  Function: 0  Alpha:   @@E24092310740 SIG2 BNSD7879 REQ1220 DSP1237 LOC 122 DAY ST BAIRNSDALE /VICTORIA ST :@BAIRNSDALE PUBLIC HOSPITAL SVVB SE 8501 E11 CC: IHTAIR2 - AIR AMBULANCE TRANSFER ACUITY: MEDIUM<NUL>'


@pytest.fixture
def sample_json_data():
    return '{"demod_name":"POCSAG1200","address":1920312,"function":3,"alpha":"Time Critical Incident - Clear ASAP - or advise Comms of Time to Clear (Via Radio)"}'


@pytest.fixture(params=FLEX_LINES.splitlines())
def flex_lines(request):
    yield request.param


def test_POCSAG_parse_line(sample_data):
    parse_line = ParseLine()
    result, json_detected = parse_line.parse(sample_data)
    assert result.address == '1622020'
    assert result.timestamp == moment.date(2024, 9, 23, 12, 38, 00)
    assert (
        result.trim_message
        == '@@E24092310740 SIG2 BNSD7879 REQ1220 DSP1237 LOC 122 DAY ST BAIRNSDALE /VICTORIA ST :@BAIRNSDALE PUBLIC HOSPITAL SVVB SE 8501 E11 CC: IHTAIR2 - AIR AMBULANCE TRANSFER ACUITY: MEDIUM'
    )
    assert json_detected is False


def test_parse_line_json(sample_json_data):
    parse_line = ParseLine()
    result, json_detected = parse_line.parse(sample_json_data)
    assert result.address == '1920312'
    assert result.trim_message == 'Time Critical Incident - Clear ASAP - or advise Comms of Time to Clear (Via Radio)'
    assert json_detected is True


def test_parse_line_invalid_not_json():
    parse_line = ParseLine(json_detected=True)
    result, json_detected = parse_line.parse('Jibberish that is not JSON')
    assert result.address == ''
    assert result.trim_message == 'ERROR: multimon-ng returned non-JSON: Jibberish that is not JSON'
    assert json_detected is True


@pytest.mark.skip(reason='Untestable for now')
def test_POCSAG_handling_numeric_message():
    line = 'Numeric Message 1234'
    parse_line = ParseLine()
    timestamp, address, message, trim_message, json_detected = parse_line.parse(line)
    assert timestamp is None
    assert address == ''
    assert message == '1234'
    assert trim_message == '1234'


@pytest.mark.skip(reason='Untestable for now')
def test_FLEX_handling_fragmented_message():
    line1 = 'FLEX [123] Fragmented Message 01/2022 F/ ALN ...[ |]'
    line2 = 'FLEX [123] Continued Message 02/2022 C/'
    parse_line = ParseLine()
    fragment_address = None
    for line in [line1, line2]:
        timestamp, address, message, trim_message = parse_line.parse(line)
        if not fragment_address:
            fragment_address = address
            assert address == '123'
            assert message is None  # First message should be empty
            continue

        assert timestamp is None
        assert address == ''
        assert message == frag[fragment_address]
        del frag[fragment_address]


@pytest.mark.skip(reason='Untestable for now')
def test_FLEX_handling_complete_message():
    line = 'FLEX [123] Complete Message 2022-09-01 12:34:56'
    parse_line = ParseLine()
    timestamp, address, message, trim_message = parse_line.parse(line)
    assert timestamp == Moment.date('2022-09-01 12:34:56', 'YYYY-MM-DD HH:mm:ss')
    assert address == ''
    assert message == ''
    assert trim_message is None


def test_default_case():
    line = 'Invalid Message'
    parse_line = ParseLine()
    result = PocsagMessage()
    result, json_detected = parse_line.parse(line)
    # assert timestamp is None
    assert result.address == ''
    assert result.trim_message == None


@pytest.mark.skip(reason='Untestable for now')
def test_FLEX_parse_line(flex_lines):
    """Test for GitHub issue https://github.com/lingfish/mmng-ui/issues/3"""

    parse_line = ParseLine()
    # result, json_detected = parse_line.parse('FLEX|2024-12-08 21:18:46|3200/4/C/C|04.063|4294942723|ALN|nsult LOPEZ PEREZ 31373343 h/o spinal compression fx\'s MRI 12/7 subacute compression deformities, no evidence of cord compression. Pt neurologically intact. any acute surgical intervention? - Donald Thommes 6314176868 [68]3fL')
    result, json_detected = parse_line.parse(flex_lines)
    # assert result.address == '4294942723'
    # assert result.timestamp == moment.date(2024, 12, 8, 21, 18, 46)
    print(result)
    assert (
            result.trim_message
            == 'nsult LOPEZ PEREZ 31373343 h/o spinal compression fx\'s MRI 12/7 subacute compression deformities, no evidence of cord compression. Pt neurologically intact. any acute surgical intervention? - Donald Thommes 6314176868 [68]3fL'
    )
    assert json_detected is False




from __future__ import annotations

import json
import re
from dataclasses import dataclass, field
from json import JSONDecodeError
from typing import Any

import moment
from moment import Moment

frag = {}


@dataclass
class PocsagMessage:
    current_time: Moment = field(default_factory=moment.now)
    timestamp: Moment = None
    address: str = None
    trim_message: str = None


@dataclass
class ParseLine:
    """
    Parse a line from multimon-ng, converted from https://github.com/pagermon/pagermon/blob/master/client/reader.js

    :param line: The multimon-ng line to parse
    :type line: str
    :param send_function_code: Not entirely sure!
    :type send_function_code: bool
    :param use_timestamp: Attempt to decode and return timestamps
    :type use_timestamp: Moment
    :return: tuple of current time, timestamp, pager address and message
    :rtype: tuple
    """
    send_function_code: bool = True
    use_timestamp: bool = True
    json_detected: bool | None = None

    def parse(self, line: str) -> tuple[PocsagMessage, bool]:
        result = PocsagMessage()
        message: str | None = None
        trim_message: str | None = None
        time_string: str | None = None
        json_line: Any | None = None
        timestamp: Moment = None

        if self.json_detected is None:
            try:
                json_line = json.loads(line)
                self.json_detected = True
            except json.JSONDecodeError:
                self.json_detected = False

        if self.json_detected:
            if not json_line:
                try:
                    json_line = json.loads(line)
                except JSONDecodeError:
                    result.trim_message = f'ERROR: multimon-ng returned non-JSON: {line}'
                    result.address = ''
                    return result, self.json_detected
            result.trim_message = re.sub(r'<[A-Za-z]{3}>', '', json_line.get('alpha', '')).replace('Ä', '[').replace('Ü', ']').strip() or ''
            result.address = str(json_line['address']) or ''

            return result, self.json_detected

        else:
            # POCSAG handling
            if re.search(r'POCSAG\d+: Address:', line):
                address_match = re.search(r'POCSAG\d+: Address:(.*?)Function', line)
                if address_match:
                    address = address_match.group(1).strip()
                    if self.send_function_code:
                        function_code_match = re.search(r'Function: (\d)', line)
                        if function_code_match:
                            address += function_code_match.group(1)

                # Handling Alpha messages
                if 'Alpha:' in line:
                    message_match = re.search(r'Alpha:(.*?)$', line)
                    if message_match:
                        message = message_match.group(1).strip()

                        if self.use_timestamp:
                            # Handle date formats like 'DD MMM YYYY HH:mm:ss'
                            if re.search(r'\d{2} \w+ \d{4} \d{2}:\d{2}:\d{2}', line):
                                time_string_match = re.search(r'\d+ \w+ \d+ \d{2}:\d{2}:\d{2}', line)
                                if time_string_match:
                                    time_string = time_string_match.group()
                                    try:
                                        timestamp = moment.date(time_string, '%d %B %Y %H:%M:%S')
                                        message = message.replace(time_string, '').strip()
                                    except ValueError:
                                        raise
                                        # pass

                            # Handle date formats like 'YYYY-MM-DD HH:mm:ss'
                            elif re.search(r'\d+-\d+-\d+ \d{2}:\d{2}:\d{2}', line):
                                time_string_match = re.search(r'\d+-\d+-\d+ \d{2}:\d{2}:\d{2}', line)
                                if time_string_match:
                                    time_string = time_string_match.group()
                                    try:
                                        timestamp = moment.date(time_string, 'YYYY-MM-DD HH:mm:ss')
                                        message = message.replace(time_string, '').strip()
                                    except ValueError:
                                        raise
                                        # pass

                        # Clean the message
                        trim_message = re.sub(r'<[A-Za-z]{3}>', '', message).replace('Ä', '[').replace('Ü', ']').strip()

                # Handling Numeric messages
                elif 'Numeric:' in line:
                    message_match = re.search(r'Numeric:(.*?)$', line)
                    if message_match:
                        message = message_match.group(1).strip()
                        trim_message = re.sub(r'<[A-Za-z]{3}>', '', message).replace('Ä', '[').replace('Ü', ']').strip()

            # FLEX handling
            elif re.search(r'FLEX[:|]', line):
                address_match = re.search(r'FLEX[:|] ?.*?[\[|](\d*?)[\]| ]', line)
                if address_match:
                    address = address_match.group(1).strip()

                # Handle timestamps within FLEX
                if self.use_timestamp:
                    if re.search(r'FLEX[:|] ?\d{2} \w+ \d{4} \d{2}:\d{2}:\d{2}', line):
                        time_string_match = re.search(r'\d+ \w+ \d+ \d{2}:\d{2}:\d{2}', line)
                        if time_string_match:
                            time_string = time_string_match.group()
                            try:
                                timestamp = moment.date(time_string, '%d %B %Y %H:%M:%S')
                            except ValueError:
                                pass

                    elif re.search(r'FLEX[:|] ?\d+-\d+-\d+ \d{2}:\d{2}:\d{2}', line):
                        time_string_match = re.search(r'\d+-\d+-\d+ \d{2}:\d{2}:\d{2}', line)
                        if time_string_match:
                            time_string = time_string_match.group()
                            try:
                                timestamp = moment.date(time_string, 'YYYY-MM-DD HH:mm:ss')
                            except ValueError:
                                pass

                # Handle fragmented and complete messages
                if re.search(r'([ |]ALN[ |]|[ |]GPN[ |]|[ |]NUM[ |])', line):
                    message_match = re.search(r'FLEX[:|].*[|\[][0-9 ]*[|\]] ?...[ |](.+)', line)
                    if message_match:
                        message = message_match.group(1).strip()

                        # Message fragmentation handling
                        if re.search(r'[ |][0-9]{4}\/[0-9]\/F\/.[ |]', line):
                            frag[address] = message
                            trim_message = ''
                        elif re.search(r'[ |][0-9]{4}\/[0-9]\/C\/.[ |]', line):
                            try:
                                trim_message = frag[address] + message
                                del frag[address]
                            except KeyError:
                                # We've seen a fragment without the starting frag, treat as final
                                trim_message = message
                        else:
                            trim_message = message

            # Default case
            else:
                address = ''
                message = None
                trim_message = ''

            result.timestamp = timestamp or None
            result.address = str(address)
            result.trim_message = trim_message or None
            return result, self.json_detected
